#include "avq7760.h"
#include <utility.h>
#include <userint.h>

/*****************************************************************************/
/*   Copyright 1998 National Instruments Corporation.  All Rights Reserved.  */
/*****************************************************************************/


/*===================== Q7760 Optical Network Analyzer ======================*/
/* LabWindows/CVI 5.0 Instrument Driver										 */
/* Original Release: June/01/2001											 */
/* By: Hitoshi Takahashi													 */
/*					 Supervisor: Yohei Hirakoso								 */
/*                   SPSpj Gyoda R&D,  ADVANTEST CORPORATION				 */
/*                   Home Page; http://www.advantest.co.jp					 */
/*                   e-mail;    info@advantest.co.jp						 */
/*---------------------------------------------------------------------------*/
/* Modification History: 													 */
/*         Rev 1.0 :  June/01/2001											 */
/*																			 */
/*===========================================================================*/
 

#include <visa.h>
#include <ansi_c.h>
#include <formatio.h>
#include "avq7760.h"

#define AVQ7760_REVISION   "Rev 1.0, 02/01, CVI 5.0" /*  Instrument driver revision */
#define BUFFER_SIZE         1024L 					  /* File I/O buffer size */
    

/*****************************************************************************/
/*= INSTRUMENT-DEPENDENT COMMAND ARRAYS =====================================*/
/*****************************************************************************/

static ViString scaleConvArr[] = {"PKM 1", "RKM 1", "PKM 0; RKM 0"};
static ViString saveRefArr[] = {"SAR", "SRT"};
static ViString calibArr[] = {"POL", "PDC", "PPN"};
static ViString xAxisArr[] = {"CEN", "SPA", "STA", "STO"};
static ViString xUnitArr[] = {"NM", "THZ"};
static ViString expDispArr[] = {"REX", "EEX"};

/*****************************************************************************/
/*= SET CONTROLLER CONDITIONS DURING READING PROCESS ========================*/
/*****************************************************************************/

ViInt16 headCntrl;
ViInt16 sdlCntrl;

/* Memorize Current Conditions and Set Header Off and Data Separator Comma  */
ViStatus CVIFUNC setCntrl (ViSession instrSession)
{
	ViStatus avq7760_status = VI_SUCCESS;
	ViChar rdBuffer[BUFFER_SIZE];    
    ViUInt32 retCount = 0;
    
    if ((avq7760_status = viWrite (instrSession, "HED?", 4, &retCount)) < 0)                
	    return avq7760_status;                                                            
	if ((avq7760_status = viRead (instrSession, rdBuffer, BUFFER_SIZE, &retCount)) < 0)        
	    return avq7760_status;                                                            
	if (Scan (rdBuffer, "%s>%d[b2]", &headCntrl) < 1)
		if (Scan (rdBuffer, "%s>HED%d[b2]", &headCntrl) < 1)
	    	return VI_ERROR_INSTR_INTERPRETING_RESPONSE;
	    
	if ((avq7760_status = viWrite (instrSession, "SDL?", 4, &retCount)) < 0)                
	    return avq7760_status;                                                            
	if ((avq7760_status = viRead (instrSession, rdBuffer, BUFFER_SIZE, &retCount)) < 0)        
	    return avq7760_status;                                                            
	if (Scan (rdBuffer, "%s>%d[b2]", &sdlCntrl) < 1)
		if (Scan (rdBuffer, "%s>SDL%d[b2]", &sdlCntrl) < 1)
	    	return VI_ERROR_INSTR_INTERPRETING_RESPONSE;
	    
	if ((avq7760_status = viWrite (instrSession, "HED 0; SDL 0", 12, &retCount)) < 0)
        return avq7760_status;
    
    return avq7760_status;    
}

/* Reset GPIB Conditions */
ViStatus CVIFUNC resetCntrl (ViSession instrSession)
{
	ViStatus avq7760_status = VI_SUCCESS;
	ViChar wrtBuffer[BUFFER_SIZE];    
    ViUInt32 retCount = 0;
    
    Fmt (wrtBuffer, "%s<HED %d; SDL %d", headCntrl, sdlCntrl);
    
    if ((avq7760_status = viWrite (instrSession, wrtBuffer, NumFmtdBytes(), &retCount)) < 0)
        return avq7760_status;
    
    return avq7760_status; 
} 

/*****************************************************************************/
/*= INSTRUMENT-DEPENDENT STATUS/RANGE STRUCTURE  ============================*/
/*****************************************************************************/
/* avq7760_stringValPair is used in the avq7760_errorMessage function        */
/*===========================================================================*/

#define MagDisplay	   0
#define GrpDlyDisplay  2
#define CDDisplay	   3
#define CdSlpDisplay   4
#define PMDDisplay	   5

#define LogScale	   0
#define LinScale	   1

typedef struct {
	ViStatus stringVal;
	ViString stringName;
} avq7760_tStringValPair;

typedef struct {
	ViInt16 displayMode;
	ViInt16 levelScale;
} avq7760_status;

avq7760_status  instrStatus; 

/*****************************************************************************/
/*= UTILITY ROUTINE DECLARATIONS (Non-Exportable Functions) =================*/
/*****************************************************************************/
ViBoolean avq7760_invalidViBooleanRange (ViBoolean val);
ViBoolean avq7760_invalidViInt16Range (ViInt16 val, ViInt16 min, ViInt16 max);
ViBoolean avq7760_invalidViInt32Range (ViInt32 val, ViInt32 min, ViInt32 max);
ViBoolean avq7760_invalidViUInt8Range (ViUInt8 val, ViUInt8 min, ViUInt8 max);
ViBoolean avq7760_invalidViUInt16Range (ViUInt16 val, ViUInt16 min, ViUInt16 max);
ViBoolean avq7760_invalidViUInt32Range (ViUInt32 val, ViUInt32 min, ViUInt32 max);
ViBoolean avq7760_invalidViReal32Range (ViReal32 val, ViReal32 min, ViReal32 max);
ViBoolean avq7760_invalidViReal64Range (ViReal64 val, ViReal64 min, ViReal64 max);
ViStatus  avq7760_initCleanUp (ViSession openRMSession, ViPSession openInstrSession,
								ViStatus currentStatus);

/*****************************************************************************/
/*====== USER-CALLABLE FUNCTIONS (Exportable Functions) =====================*/
/*****************************************************************************/

/*===========================================================================*/
/* Function: Initialize                                                      */
/* Purpose:  This function opens the instrument, queries the instrument      */
/*           for its ID, and initializes the instrument to a known state.    */
/*===========================================================================*/
ViStatus _VI_FUNC avq7760_Initialize (ViRsrc resourceName, ViBoolean IDQuery,
                                       ViBoolean resetDevice, ViSession *instrSession)
{
    ViStatus avq7760_status = VI_SUCCESS;
    ViSession rmSession = 0;
    ViUInt32 retCnt = 0;
    ViByte rdBuffer[BUFFER_SIZE];

    /*- Check input parameter ranges ----------------------------------------*/
    if (avq7760_invalidViBooleanRange (IDQuery))
        return VI_ERROR_PARAMETER2;
    if (avq7760_invalidViBooleanRange (resetDevice))
        return VI_ERROR_PARAMETER3;

    /*- Open instrument session ---------------------------------------------*/
    if ((avq7760_status = viOpenDefaultRM (&rmSession)) < 0)
        return avq7760_status;

    if ((avq7760_status = viOpen (rmSession, resourceName, VI_NULL, VI_NULL, instrSession)) < 0) {
        viClose (rmSession);
        return avq7760_status;
    }

    /*- Configure VISA Formatted I/O ----------------------------------------*/
    if ((avq7760_status = viSetAttribute (*instrSession, VI_ATTR_TMO_VALUE, 10000)) < 0)
            return avq7760_initCleanUp (rmSession, instrSession, avq7760_status);
  
    if ((avq7760_status = viSetBuf (*instrSession, VI_READ_BUF|VI_WRITE_BUF, 4000)) < 0)
            return avq7760_initCleanUp (rmSession, instrSession, avq7760_status);
  
    if ((avq7760_status = viSetAttribute (*instrSession, VI_ATTR_WR_BUF_OPER_MODE,
                            VI_FLUSH_ON_ACCESS)) < 0)
            return avq7760_initCleanUp (rmSession, instrSession, avq7760_status);
    
    if ((avq7760_status = viSetAttribute (*instrSession, VI_ATTR_RD_BUF_OPER_MODE,
                            VI_FLUSH_ON_ACCESS)) < 0)
            return avq7760_initCleanUp (rmSession, instrSession, avq7760_status);

    /*- Identification Query ------------------------------------------------*/
    if (IDQuery) 
    {
        if ((avq7760_status = viWrite (*instrSession, "*IDN?", 5, &retCnt)) < 0)
            return avq7760_initCleanUp (rmSession, instrSession, avq7760_status);
        if ((avq7760_status = viRead (*instrSession, rdBuffer, BUFFER_SIZE, &retCnt)) < 0)
            return avq7760_status;

        Scan (rdBuffer, "ADVANTEST,Q7760");
        if (NumFmtdBytes () != 15) 
          	return avq7760_initCleanUp (rmSession, instrSession, VI_ERROR_FAIL_ID_QUERY);
    }
        
    /*- Reset instrument ----------------------------------------------------*/
    if (resetDevice)
    {
        if ((avq7760_status = avq7760_Reset (*instrSession)) < 0)
            return avq7760_initCleanUp (rmSession, instrSession, avq7760_status);
    }       
          
    return avq7760_status;
}

/*===========================================================================*/
/* Function: Set Function                                                    */
/* Purpose:  This function sets the measurement mode.						 */
/*===========================================================================*/
ViStatus _VI_FUNC avq7760_Set_Function (ViSession instrSession,
                                        ViBoolean msMode, ViInt16 msFrmt,
                                        ViBoolean lvlDisp)
{
	ViStatus avq7760_status = VI_SUCCESS;
	ViChar wrtBuffer[BUFFER_SIZE];
    ViUInt32 retCount = 0; 
    ViUInt32 total = 0;
    
    if (avq7760_invalidViBooleanRange (msMode))
        return VI_ERROR_PARAMETER2;
    if (avq7760_invalidViInt16Range (msFrmt, 0, 5))
        return VI_ERROR_PARAMETER3;
    if (msFrmt==1)
   		return VI_ERROR_PARAMETER3;
    if (avq7760_invalidViBooleanRange (lvlDisp))
        return VI_ERROR_PARAMETER4; 
        
    Fmt (wrtBuffer, "%s<RTR %d; FOR %d", msMode, msFrmt);
    total += NumFmtdBytes();
    
    instrStatus.displayMode = msFrmt;
    
    if (msFrmt==0)
    {
    	Fmt (wrtBuffer, "%s[a]<; LIN %d", lvlDisp);
    	total += NumFmtdBytes();
    	
    	instrStatus.levelScale = lvlDisp;
    }
    
    if ((avq7760_status = viWrite (instrSession, wrtBuffer, total, &retCount)) < 0)
        return avq7760_status;
    
    return avq7760_status;
}

/*===========================================================================*/
/* Function: Set Advance                                                     */
/* Purpose:  This function sets fiber index, lambda patrameters and free	 */
/*			 rotate.														 */
/*===========================================================================*/
ViStatus _VI_FUNC avq7760_Set_Advance (ViSession instrSession,
                                       ViReal64 fbrIndx, ViBoolean atLmbdLen,
                                       ViReal64 lmbdLen, ViReal64 qdLmbd, 
                                       ViReal64 hfLmbd, ViBoolean rotate,
                                       ViBoolean calFbrLen)
{
	ViStatus avq7760_status = VI_SUCCESS;
	ViChar wrtBuffer[BUFFER_SIZE];
    ViUInt32 retCount = 0; 
    ViUInt32 total = 0;
    
    if (avq7760_invalidViReal64Range (fbrIndx, 1.0, 2.0))
        return VI_ERROR_PARAMETER2;
    if (avq7760_invalidViBooleanRange (atLmbdLen))
        return VI_ERROR_PARAMETER3;
    if (avq7760_invalidViReal64Range (lmbdLen, 1525.0, 1635.0))
        return VI_ERROR_PARAMETER4;
    if (avq7760_invalidViReal64Range (qdLmbd, -360.0, 360.0))
        return VI_ERROR_PARAMETER5;
    if (avq7760_invalidViReal64Range (hfLmbd, -360.0, 360.0))
        return VI_ERROR_PARAMETER6;    
    if (avq7760_invalidViBooleanRange (rotate))
        return VI_ERROR_PARAMETER7;
    if (avq7760_invalidViBooleanRange (calFbrLen))
        return VI_ERROR_PARAMETER8;
        
        
    Fmt (wrtBuffer, "%s<IND %f; ALD %d; LWD %fNM; QWP %f; HWP %f; RDM %d",
    	 fbrIndx, atLmbdLen, lmbdLen, qdLmbd, hfLmbd, rotate);
    total += NumFmtdBytes();
    
    if (calFbrLen==VI_ON)
    {
    	Fmt (wrtBuffer, "%s[a]<; ELG");
    	total += NumFmtdBytes();
    }
        
    if ((avq7760_status = viWrite (instrSession, wrtBuffer, total, &retCount)) < 0)
        return avq7760_status;
    
    return avq7760_status;    
}

/*===========================================================================*/
/* Function: Set Domain			                                             */
/* Purpose:  This function sets the domain									 */
/*===========================================================================*/
ViStatus _VI_FUNC avq7760_Set_Domain (ViSession instrSession,
                                      ViBoolean domain)
{
	ViStatus avq7760_status = VI_SUCCESS;
	ViChar wrtBuffer[BUFFER_SIZE];
    ViUInt32 retCount = 0; 
    
    if (avq7760_invalidViBooleanRange (domain))
        return VI_ERROR_PARAMETER2;

    Fmt (wrtBuffer, "%s<FDO %d", domain);
    
    if ((avq7760_status = viWrite (instrSession, wrtBuffer, NumFmtdBytes(), &retCount)) < 0)
        return avq7760_status;
    
    return avq7760_status;    
}

/*===========================================================================*/
/* Function: Set Center and Span                                             */
/* Purpose:  This function sets the horizontal axis's display range of the   */
/*			 graph by center and span.										 */
/*===========================================================================*/
ViStatus _VI_FUNC avq7760_Set_CenterSpan (ViSession instrSession, 
										  ViBoolean domain,
                                          ViReal64 center, ViReal64 span)
{
	ViStatus avq7760_status = VI_SUCCESS;
	ViChar wrtBuffer[BUFFER_SIZE];
    ViUInt32 retCount = 0; 
    
    if (avq7760_invalidViBooleanRange (domain))
        return VI_ERROR_PARAMETER2;

    if (domain==0)   /* Wavelength */
    {
    	if (avq7760_invalidViReal64Range (center, 1525.0, 1635.0))
        	return VI_ERROR_PARAMETER3;
		if (avq7760_invalidViReal64Range (span, 0.0, 110.0))
        	return VI_ERROR_PARAMETER4;
    }
    
    else			 /* Frequency */
    {
    	if (avq7760_invalidViReal64Range (center, 183.3593, 196.5852))
        	return VI_ERROR_PARAMETER3;
		if (avq7760_invalidViReal64Range (span, 0.0, 13.2259))
        	return VI_ERROR_PARAMETER4;
    }

    if (domain==0)     /* Wavelength */
    	Fmt (wrtBuffer, "%s<FDO %d; CEN %fNM; SPA %fNM", domain, center, span);
    else			   /* Frequency */
    	Fmt (wrtBuffer, "%s<FDO %d; CEN %fTHZ; SPA %fTHZ", domain, center, span);
    
    if ((avq7760_status = viWrite (instrSession, wrtBuffer, NumFmtdBytes(), &retCount)) < 0)
        return avq7760_status;
    
    return avq7760_status;    
}

/*===========================================================================*/
/* Function: Set Start and Stop                                              */
/* Purpose:  This function sets the horizontal axis's display range of the   */
/*			 graph by start and stop.										 */
/*===========================================================================*/
ViStatus _VI_FUNC avq7760_Set_StartStop (ViSession instrSession,
										 ViBoolean domain,
                                         ViReal64 start, ViReal64 stop)
{
	ViStatus avq7760_status = VI_SUCCESS;
	ViChar wrtBuffer[BUFFER_SIZE];
    ViUInt32 retCount = 0;
    
    if (avq7760_invalidViBooleanRange (domain))
        return VI_ERROR_PARAMETER2;

    if (domain==0)	/* Wavelength */
    {
    	if (avq7760_invalidViReal64Range (start, 1525.0, 1635.0))
        	return VI_ERROR_PARAMETER3;
		if (avq7760_invalidViReal64Range (stop, 1525.0, 1635.0))
        	return VI_ERROR_PARAMETER4;
    }
    
    else			/* Frequency */
    {
    	if (avq7760_invalidViReal64Range (start, 183.3593, 196.5852))
        	return VI_ERROR_PARAMETER3;
		if (avq7760_invalidViReal64Range (stop, 183.3593, 196.5852))
        	return VI_ERROR_PARAMETER4;
    }
    
    if (domain==0)     /* Wavelength */
    	Fmt (wrtBuffer, "%s<FDO %d; STA %fNM; STO %fNM", domain, start, stop);
    else			   /* Frequency */
    	Fmt (wrtBuffer, "%s<FDO %d; STA %fTHZ; STO %fTHZ", domain, start, stop);
    
    if ((avq7760_status = viWrite (instrSession, wrtBuffer, NumFmtdBytes(), &retCount)) < 0)
        return avq7760_status;
    
    return avq7760_status;    
}

/*===========================================================================*/
/* Function: Set Wavelength or frequency	                                 */
/* Purpose:  This function selects center, span, start or stop, and sets its */
/*			 value. This funtion does not select X aixs mode. Instead, it    */
/*			 sets X axis unit.												 */
/*===========================================================================*/
ViStatus CVIFUNC avq7760_Set_WavelenFreq (ViSession instrSession,
                                          ViInt16 setMode, ViReal64 value,
                                          ViBoolean unit)
{
	ViStatus avq7760_status = VI_SUCCESS;
	ViChar wrtBuffer[BUFFER_SIZE];
    ViUInt32 retCount = 0; 
    
    if (avq7760_invalidViInt16Range (setMode, 0, 3))
        return VI_ERROR_PARAMETER2;
    if (avq7760_invalidViBooleanRange (unit))
        return VI_ERROR_PARAMETER4;
        
    Fmt (wrtBuffer, "%s<%s %f%s", xAxisArr[setMode], value, xUnitArr[unit]);    
    
    if ((avq7760_status = viWrite (instrSession, wrtBuffer, NumFmtdBytes(), &retCount)) < 0)
        return avq7760_status;
    
    return avq7760_status;    
}

/*===========================================================================*/
/* Function: Expansion Display	                                             */
/* Purpose:  This function switches the expansion display.					 */
/*===========================================================================*/
ViStatus CVIFUNC avq7760_Set_ExpDisplay (ViSession instrSession, ViBoolean expDisp)
{
	ViStatus avq7760_status = VI_SUCCESS;
	ViChar wrtBuffer[BUFFER_SIZE];
    ViUInt32 retCount = 0; 
    
    if (avq7760_invalidViBooleanRange (expDisp))
        return VI_ERROR_PARAMETER2;
        
    Fmt (wrtBuffer, "%s<%s", expDispArr[expDisp]);    
    
    if ((avq7760_status = viWrite (instrSession, wrtBuffer, NumFmtdBytes(), &retCount)) < 0)
        return avq7760_status;
    
    return avq7760_status;    
}

/*===========================================================================*/
/* Function: Set Cursor to Center	                                         */
/* Purpose:  When only one cursor X1 or X2 is displayed, displayed cursor's  */
/*			 X value is set to center. When both cursors are displayed, 	 */
/*			 middle X value is set to center.								 */
/*===========================================================================*/
ViStatus CVIFUNC avq7760_Set_CursorCenter (ViSession instrSession)
{
	ViStatus avq7760_status = VI_SUCCESS;
    ViUInt32 retCount = 0; 
    
    if ((avq7760_status = viWrite (instrSession, "CUC", 3, &retCount)) < 0)
        return avq7760_status;
    
    return avq7760_status;    
}

/*===========================================================================*/
/* Function: Set Level Scale	                                             */
/* Purpose:  This function sets the display method of the level.			 */
/*===========================================================================*/
ViStatus _VI_FUNC avq7760_Set_LevelScale (ViSession instrSession,
                                          ViBoolean atScl, ViInt16 lvScl,
                                          ViReal64 refLvl)
{
	ViStatus avq7760_status = VI_SUCCESS;
	ViChar wrtBuffer[BUFFER_SIZE];
    ViUInt32 retCount = 0; 
    
    if (avq7760_invalidViBooleanRange (atScl))
        return VI_ERROR_PARAMETER2;
    if (avq7760_invalidViInt16Range (lvScl, 0, 5))
        return VI_ERROR_PARAMETER5;
        
    if (atScl==0)   /* Auto Scale Off */
    	if (instrStatus.displayMode==MagDisplay && instrStatus.levelScale==LogScale)
    		Fmt (wrtBuffer, "%s<LEV %d; REF %fDB", lvScl, refLvl);
    	else
    		Fmt (wrtBuffer, "%s<LEV %d; REF %fN", lvScl, refLvl*1.0e9); 
    else			/* Auto Scale On */
    	Fmt (wrtBuffer, "%s<LAU");
    	
    if ((avq7760_status = viWrite (instrSession, wrtBuffer, NumFmtdBytes(), &retCount)) < 0)
        return avq7760_status;
    
    return avq7760_status;  	
}

/*===========================================================================*/
/* Function: Set Scale Conversion                                            */
/* Purpose:  This function sets the scale conversion.						 */
/*===========================================================================*/
ViStatus _VI_FUNC avq7760_Set_ScaleConversion (ViSession instrSession,
                                               ViInt16 sclConv, ViReal64 fiberLen)
{
	ViStatus avq7760_status = VI_SUCCESS;
	ViChar wrtBuffer[BUFFER_SIZE];
    ViUInt32 retCount = 0; 
    
    if (avq7760_invalidViInt16Range (sclConv, 0, 2))
        return VI_ERROR_PARAMETER2;
    if (avq7760_invalidViReal64Range (fiberLen, 0.0, 99999.0))
        return VI_ERROR_PARAMETER3;
        
    Fmt (wrtBuffer, "%s<FIB %f; %s", fiberLen, scaleConvArr[sclConv]);
    
    if ((avq7760_status = viWrite (instrSession, wrtBuffer, NumFmtdBytes(), &retCount)) < 0)
        return avq7760_status;
    
    return avq7760_status;  
}

/*===========================================================================*/
/* Function: Set Modulation Frequency                                        */
/* Purpose:  This function sets the modulation frequency.					 */
/*===========================================================================*/
ViStatus _VI_FUNC avq7760_Set_ModulationFreq (ViSession instrSession,
                                              ViBoolean atMdlFreq, ViReal64 mdlFreq)
{
	ViStatus avq7760_status = VI_SUCCESS;
	ViChar wrtBuffer[BUFFER_SIZE];
    ViUInt32 retCount = 0; 
    
    if (avq7760_invalidViBooleanRange (atMdlFreq))
        return VI_ERROR_PARAMETER2;
    if (avq7760_invalidViReal64Range (mdlFreq, 0.04, 3.0))
        return VI_ERROR_PARAMETER3;
    
    if (atMdlFreq==0)   /* Auto Off */
    	Fmt (wrtBuffer, "%s<AUM 0; MOF %fGHZ", mdlFreq);
    else				/* Auto On */
    	Fmt (wrtBuffer, "%s<AUM 1");
    	
    if ((avq7760_status = viWrite (instrSession, wrtBuffer, NumFmtdBytes(), &retCount)) < 0)
        return avq7760_status;
    
    return avq7760_status;  	
}

/*===========================================================================*/
/* Function: Set Display			                                         */
/* Purpose:  This function sets the display menthod of the instrument's		 */
/*			 screen.														 */
/*===========================================================================*/
ViStatus _VI_FUNC avq7760_Set_Display (ViSession instrSession,
                                       ViBoolean dual, ViBoolean impose, 
                                       ViBoolean actUL, ViBoolean actLR,
                                       ViBoolean grid, ViBoolean bothMs,
                                       ViBoolean exchange,
                                       ViBoolean paste, ViBoolean erase)
{
	ViStatus avq7760_status = VI_SUCCESS;
	ViChar wrtBuffer[BUFFER_SIZE];
    ViUInt32 retCount = 0;
    ViUInt32 total = 0;
    
    if (avq7760_invalidViBooleanRange (dual))
        return VI_ERROR_PARAMETER2;    
    if (avq7760_invalidViBooleanRange (impose))
        return VI_ERROR_PARAMETER3;
    if (dual==1 && impose==1)
    	return VI_ERROR_PARAMETER3;
    if (avq7760_invalidViBooleanRange (actUL))
        return VI_ERROR_PARAMETER4;    
    if (avq7760_invalidViBooleanRange (actLR))
        return VI_ERROR_PARAMETER5;    
    if (avq7760_invalidViBooleanRange (grid))
        return VI_ERROR_PARAMETER6;    
    if (avq7760_invalidViBooleanRange (bothMs))
        return VI_ERROR_PARAMETER7;
    if (avq7760_invalidViBooleanRange (exchange))
        return VI_ERROR_PARAMETER8;    
    if (avq7760_invalidViBooleanRange (paste))
        return VI_ERROR_INSTR_PARAMETER9;    
    if (avq7760_invalidViBooleanRange (erase))
        return VI_ERROR_INSTR_PARAMETER10;
        
    Fmt (wrtBuffer, "%s<GRI %d", grid);
    total += NumFmtdBytes();
    
    if (dual==0)   /* Dual Display Off */
    	Fmt (wrtBuffer, "%s[a]<; DUA 0");
    else		   /* Dual Display On */
    	Fmt (wrtBuffer, "%s[a]<; DUA 1; AUL %d", actUL);
    total += NumFmtdBytes();
    
    if (impose==0)   /* Superimpose Off */
    	Fmt (wrtBuffer, "%s[a]<; SIM 0");
    else			 /* Superimpose On */
    	Fmt (wrtBuffer, "%s[a]<; SIM 1; AUL %d", actLR);
    total += NumFmtdBytes();
    
    if (dual==1 || impose==1)   /* Dual Display or Superimpose is On */
    {
    	Fmt (wrtBuffer, "%s[a]<; BOM %d", bothMs);
    	total += NumFmtdBytes();
    
    	if (dual==1)
    		if (exchange==1)
    		{
    			Fmt (wrtBuffer, "%s[a]<; XUL");
    			total += NumFmtdBytes();
    		}
    		
    	else
    	{
    		if (paste==1)
    		{
    			Fmt (wrtBuffer, "%s[a]<; PAD");
    			total += NumFmtdBytes();
    		}
    		
    		if (erase==1)
    		{
    			Fmt (wrtBuffer, "%s[a]<; ERD");
    			total += NumFmtdBytes();
    		}
    	}
    }
    
    if ((avq7760_status = viWrite (instrSession, wrtBuffer, total, &retCount)) < 0)
        return avq7760_status;
    
    return avq7760_status;  
}

/*===========================================================================*/
/* Function: Set Limit Line			                                         */
/* Purpose:  This function sets the limit line parameters.					 */
/*===========================================================================*/
ViStatus _VI_FUNC avq7760_Set_LimitLine (ViSession instrSession,
                                         ViBoolean lmtLine, ViInt16 pattern)
{
	ViStatus avq7760_status = VI_SUCCESS;
	ViChar wrtBuffer[BUFFER_SIZE];
    ViUInt32 retCount = 0;
    
    if (avq7760_invalidViBooleanRange (lmtLine))
        return VI_ERROR_PARAMETER2;    
    if (avq7760_invalidViInt16Range (pattern, 1, 5))
        return VI_ERROR_PARAMETER3;
        
    Fmt (wrtBuffer, "%s<LPT %d; LIM %d", pattern, lmtLine);
    
    if ((avq7760_status = viWrite (instrSession, wrtBuffer, NumFmtdBytes(), &retCount)) < 0)
        return avq7760_status;
    
    return avq7760_status;     
}

/*===========================================================================*/
/* Function: Display Report			                                         */
/* Purpose:  This function sets the parameters to display measurement		 */
/*			 results.														 */
/*===========================================================================*/
ViStatus CVIFUNC avq7760_DisplayReport (ViSession instrSession,
                                        ViBoolean dispReport, ViString comment,
                                        ViString id1, ViString id2, 
                                        ViString id3, ViString fileName, 
                                        ViReal64 start, ViReal64 stop,
                                        ViReal64 resolution, ViReal64 reference,
                                        ViBoolean exePrint)
{
	ViStatus avq7760_status = VI_SUCCESS;
	ViChar wrtBuffer[BUFFER_SIZE];
    ViUInt32 retCount = 0;
    ViUInt32 total = 0;
    
    if (avq7760_invalidViBooleanRange (dispReport))
        return VI_ERROR_PARAMETER2; 
    if (avq7760_invalidViReal64Range (start, 1525.0, 1634.9))
        return VI_ERROR_PARAMETER8;    
    if (avq7760_invalidViReal64Range (stop, 1525.1, 1635.0))
        return VI_ERROR_INSTR_PARAMETER9;    
    if (avq7760_invalidViReal64Range ((stop - start)/resolution, 10.0, 600.0))
        return VI_ERROR_INSTR_PARAMETER10;    
    if (avq7760_invalidViReal64Range (reference, 1000.0, 2000.0))
        return VI_ERROR_INSTR_PARAMETER11;    
    if (avq7760_invalidViBooleanRange (exePrint))
        return VI_ERROR_INSTR_PARAMETER12; 
        
    if (dispReport==VI_ON)
    {
    	Fmt (wrtBuffer, "%s<REP 1; RCO #%s#; RCA #%s#; RSM #%s#; RFB #%s#; RFI #%s#; RSL %fNM; RPL %fNM; RRS %fNM; RRF %fNM",
    		 comment, id1, id2, id3, fileName, start, stop, resolution, reference);
    	total += NumFmtdBytes();
    	
    	if (exePrint==VI_ON)
    	{
    		Fmt (wrtBuffer, "%s[a]<; ERP");
    		total += NumFmtdBytes();
    	}
    }
    
    else
    {
    	Fmt (wrtBuffer, "%s<REP 0");
    	total += NumFmtdBytes();
    }
        
    if ((avq7760_status = viWrite (instrSession, wrtBuffer, total, &retCount)) < 0)
        return avq7760_status;
    
	return avq7760_status;    
}

/*===========================================================================*/
/* Function: Average				                                         */
/* Purpose:  This function sets the average parameters.						 */
/*===========================================================================*/
ViStatus _VI_FUNC avq7760_Average (ViSession instrSession, 
								   ViBoolean avg, ViInt16 avgCnt)
{															 
	ViStatus avq7760_status = VI_SUCCESS;
	ViChar wrtBuffer[BUFFER_SIZE];
    ViUInt32 retCount = 0;
    
    if (avq7760_invalidViBooleanRange (avg))
        return VI_ERROR_PARAMETER2;    
    if (avq7760_invalidViInt16Range (avgCnt, 1, 16))
        return VI_ERROR_PARAMETER3;
        
    Fmt (wrtBuffer, "%s<AVG %d; EAV %d", avgCnt, avg);
    
    if ((avq7760_status = viWrite (instrSession, wrtBuffer, NumFmtdBytes(), &retCount)) < 0)
        return avq7760_status;
    
    return avq7760_status;     
}

/*===========================================================================*/
/* Function: Average				                                         */
/* Purpose:  This function sets the average parameters.						 */
/*===========================================================================*/
ViStatus _VI_FUNC avq7760_Unwrap (ViSession instrSession, 
								  ViBoolean unwrap, ViReal64 threshold)
{
	ViStatus avq7760_status = VI_SUCCESS;
	ViChar wrtBuffer[BUFFER_SIZE];
    ViUInt32 retCount = 0;
    
    if (avq7760_invalidViBooleanRange (unwrap))
        return VI_ERROR_PARAMETER2;    
    if (avq7760_invalidViReal64Range (threshold, -100.0, 20.0))
        return VI_ERROR_PARAMETER3;
        
    Fmt (wrtBuffer, "%s<UWP %f; UWR %d", threshold, unwrap);
    
    if ((avq7760_status = viWrite (instrSession, wrtBuffer, NumFmtdBytes(), &retCount)) < 0)
        return avq7760_status;
    
    return avq7760_status;     
}

/*===========================================================================*/
/* Function: Smoothing				                                         */
/* Purpose:  This function sets the smoothing parameters.					 */
/*===========================================================================*/
ViStatus _VI_FUNC avq7760_Smoothing (ViSession instrSession,
									 ViBoolean domain,
                                     ViReal64 width, ViBoolean smooth)
{
	ViStatus avq7760_status = VI_SUCCESS;
	ViChar wrtBuffer[BUFFER_SIZE];
    ViUInt32 retCount = 0; 
    
    if (avq7760_invalidViBooleanRange (domain))
        return VI_ERROR_PARAMETER2;
    if (avq7760_invalidViBooleanRange (smooth))
        return VI_ERROR_PARAMETER4;
        
    if (domain==0)     /* Wavelength */
    	Fmt (wrtBuffer, "%s<FDO 0; SMW %fNM; SMO %d", width, smooth);
    else			   /* Frequency */
    	Fmt (wrtBuffer, "%s<FDO 1; SMW %fTHZ; SMO %d", width, smooth);
    
    if ((avq7760_status = viWrite (instrSession, wrtBuffer, NumFmtdBytes(), &retCount)) < 0)
        return avq7760_status;
    
    return avq7760_status;    
}

/*===========================================================================*/
/* Function: Sweep Mode				                                         */
/* Purpose:  This function sets the sweep mode parameters.					 */
/*===========================================================================*/
ViStatus _VI_FUNC avq7760_Sweep (ViSession instrSession, 
								 ViBoolean swpMode, ViBoolean fxMode, 
                                 ViBoolean select,
                                 ViInt16 dtPnt, ViReal64 res)
{
	ViStatus avq7760_status = VI_SUCCESS;
	ViChar wrtBuffer[BUFFER_SIZE];
    ViUInt32 retCount = 0; 
    
    if (avq7760_invalidViBooleanRange (swpMode))
        return VI_ERROR_PARAMETER2;
    if (avq7760_invalidViBooleanRange (fxMode))
        return VI_ERROR_PARAMETER3;
    if (avq7760_invalidViBooleanRange (select))
        return VI_ERROR_PARAMETER4;    
    if (swpMode==0)   /* Continue Mode */
    {
    	if (avq7760_invalidViInt16Range (dtPnt, 101, 12001))
        	return VI_ERROR_PARAMETER5;
    	if (avq7760_invalidViReal64Range (res, 0.00015, 0.02))
        	return VI_ERROR_PARAMETER6;
    }
    else			  /* Step Mode */
    {
    	if (avq7760_invalidViInt16Range (dtPnt, 11, 301))
        	return VI_ERROR_PARAMETER5;
    	if (avq7760_invalidViReal64Range (res, 0.01, 11.0))
        	return VI_ERROR_PARAMETER6;
    }
    
    if (swpMode==0)   /* Continue Mode */
    	if (select==0)   /* Set Resolution */
    		Fmt (wrtBuffer, "%s<DTM 0; CRS %fNM; CRP %d", res, fxMode);
    	else			 /* Set Data Points */
    		Fmt (wrtBuffer, "%s<DTM 0; CDP %d; CRP %d", dtPnt, fxMode);
    
    else			  /* Step Mode */
    	if (select==0)   /* Set Resolution */
    		Fmt (wrtBuffer, "%s<DTM 1; STL %fNM; SRP %d", res, fxMode);
    	else			 /* Set Data Points */
    		Fmt (wrtBuffer, "%s<DTM 1; DPS %d; SRP %d", dtPnt, fxMode);
    	
    if ((avq7760_status = viWrite (instrSession, wrtBuffer, NumFmtdBytes(), &retCount)) < 0)
        return avq7760_status;
    
    return avq7760_status;    	
}

/*===========================================================================*/
/* Function: Measurement Mode		                                         */
/* Purpose:  This function sets the measurement mode.						 */
/*===========================================================================*/
ViStatus _VI_FUNC avq7760_MeasMode (ViSession instrSession, ViInt16 measMode)
{
	ViStatus avq7760_status = VI_SUCCESS;
	ViChar wrtBuffer[BUFFER_SIZE];
    ViUInt32 retCount = 0;
    
    if (measMode!=0 && measMode!=2)
    	return VI_ERROR_PARAMETER2;
    
    Fmt (wrtBuffer, "%s<DIF %d", measMode);
    
    if ((avq7760_status = viWrite (instrSession, wrtBuffer, NumFmtdBytes(), &retCount)) < 0)
        return avq7760_status;
    
    return avq7760_status;
}

/*===========================================================================*/
/* Function: Fit					                                         */
/* Purpose:  This function sets the fit parameters.							 */
/*===========================================================================*/
ViStatus _VI_FUNC avq7760_Fit (ViSession instrSession, 
							   ViBoolean fit, ViInt16 fitMode, 
							   ViBoolean dtDisp, ViBoolean dtLine,
							   ViBoolean partial)
{
	ViStatus avq7760_status = VI_SUCCESS;
	ViChar wrtBuffer[BUFFER_SIZE];
    ViUInt32 retCount = 0; 
	ViUInt32 total = 0;
    	
    if (avq7760_invalidViBooleanRange (fit))
        return VI_ERROR_PARAMETER2;
    if (avq7760_invalidViInt16Range (fitMode, 0, 3))
        return VI_ERROR_PARAMETER3;
    if (avq7760_invalidViBooleanRange (dtDisp))
        return VI_ERROR_PARAMETER4;
    if (avq7760_invalidViBooleanRange (dtLine))
        return VI_ERROR_PARAMETER5;
    if (avq7760_invalidViBooleanRange (partial))
        return VI_ERROR_PARAMETER6;
    
    if (dtDisp==0)   /* Data Display Off */
    	Fmt (wrtBuffer, "%s<MED 0");
    else			 /* Data Display On */
    	Fmt (wrtBuffer, "%s<MED 1; DOT %d", dtLine);
    total += NumFmtdBytes();
    
    Fmt (wrtBuffer, "%s[a]<; FIM %d; PFT %d; FIT %d", fitMode, partial, fit);
    total += NumFmtdBytes();
    
    if ((avq7760_status = viWrite (instrSession, wrtBuffer, total, &retCount)) < 0)
        return avq7760_status;
    
    return avq7760_status;
}

/*===========================================================================*/
/* Function: Sensitivity			                                         */
/* Purpose:  This function sets the sensitivity for the measurement.		 */
/*===========================================================================*/
ViStatus _VI_FUNC avq7760_Sensitivity (ViSession instrSession,
                                       ViInt16 sens)
{
	ViStatus avq7760_status = VI_SUCCESS;
	ViChar wrtBuffer[BUFFER_SIZE];
    ViUInt32 retCount = 0; 
	
    if (avq7760_invalidViInt16Range (sens, 0, 3))
        return VI_ERROR_PARAMETER2;
        
    Fmt (wrtBuffer, "%s<SEN %d", sens);
    
    if ((avq7760_status = viWrite (instrSession, wrtBuffer, NumFmtdBytes(), &retCount)) < 0)
        return avq7760_status;
    
    return avq7760_status;
}

/*===========================================================================*/
/* Function: Lambda Compensation	                                         */
/* Purpose:  This function sets the lambda compensation parameters.			 */
/*===========================================================================*/
ViStatus _VI_FUNC avq7760_LambdaComp (ViSession instrSession,
                                      ViBoolean cmp, ViInt16 mtrType)
{
	ViStatus avq7760_status = VI_SUCCESS;
	ViChar wrtBuffer[BUFFER_SIZE];
    ViUInt32 retCount = 0; 
	
    if (avq7760_invalidViBooleanRange (cmp))
        return VI_ERROR_PARAMETER2;
    if (avq7760_invalidViInt16Range (mtrType, 0, 2))
        return VI_ERROR_PARAMETER3;    
        
    Fmt (wrtBuffer, "%s<WMT %d; WCA %d", mtrType, cmp);
    
    if ((avq7760_status = viWrite (instrSession, wrtBuffer, NumFmtdBytes(), &retCount)) < 0)
        return avq7760_status;
    
    return avq7760_status;
}

/*===========================================================================*/
/* Function: Set Level Offset		                                         */
/* Purpose:  This function sets the level offset for reflect mode and		 */
/*			 transmit mode.													 */
/*===========================================================================*/
ViStatus _VI_FUNC avq7760_Set_LevelOffset (ViSession instrSession,
                                           ViReal64 rfLvlOfst, ViReal64 trLvlOfst)
{
	ViStatus avq7760_status = VI_SUCCESS;
	ViChar wrtBuffer[BUFFER_SIZE];
    ViUInt32 retCount = 0; 
	
    if (avq7760_invalidViReal64Range (rfLvlOfst, -20.0, 20.0))
        return VI_ERROR_PARAMETER2;
    if (avq7760_invalidViReal64Range (trLvlOfst, -20.0, 20.0))
        return VI_ERROR_PARAMETER3;    
    
    Fmt (wrtBuffer, "%s<LCA %f; LCT %f", rfLvlOfst, trLvlOfst);
    
    if ((avq7760_status = viWrite (instrSession, wrtBuffer, NumFmtdBytes(), &retCount)) < 0)
        return avq7760_status;
    
    return avq7760_status;
}

/*===========================================================================*/
/* Function: Save Reference			                                         */
/* Purpose:  This function saves measurement value as compensation data.	 */
/*===========================================================================*/
ViStatus _VI_FUNC avq7760_SaveReference (ViSession instrSession,
                                         ViBoolean saveAs)
{
	ViStatus avq7760_status = VI_SUCCESS;
	ViChar wrtBuffer[BUFFER_SIZE];
    ViUInt32 retCount = 0; 
	
    if (avq7760_invalidViBooleanRange (saveAs))
        return VI_ERROR_PARAMETER2;
        
    Fmt (wrtBuffer, "%s<%s", saveRefArr[saveAs]);    
       
    if ((avq7760_status = viWrite (instrSession, wrtBuffer, NumFmtdBytes(), &retCount)) < 0)
        return avq7760_status;
    
    return avq7760_status;   
}

/*===========================================================================*/
/* Function: Normalize				                                         */
/* Purpose:  This function selects normalized mode using compensation data.  */
/*			 The compensation data are saved in the instrument's memory.	 */
/*===========================================================================*/
ViStatus _VI_FUNC avq7760_Normalize (ViSession instrSession, 
									 ViBoolean reflect, ViBoolean transmit)
{
	ViStatus avq7760_status = VI_SUCCESS;
	ViChar wrtBuffer[BUFFER_SIZE];
    ViUInt32 retCount = 0; 
	
    if (avq7760_invalidViBooleanRange (reflect))
        return VI_ERROR_PARAMETER2;
    if (avq7760_invalidViBooleanRange (transmit))
        return VI_ERROR_PARAMETER3;    
    
    Fmt (wrtBuffer, "%s<NOR %d; NRT %d", reflect, transmit);
    
    if ((avq7760_status = viWrite (instrSession, wrtBuffer, NumFmtdBytes(), &retCount)) < 0)
        return avq7760_status;
    
    return avq7760_status;   
}

/*===========================================================================*/
/* Function: Execute Calibration	                                         */
/* Purpose:  This function executes POL Calib (calibration of polarization   */
/*			 controller), PMD Normalize (polarization mode dispersion		 */
/*			 normalization), or both.										 */
/*===========================================================================*/
ViStatus _VI_FUNC avq7760_Exe_Calibration (ViSession instrSession, ViInt16 calib)
{
	ViStatus avq7760_status = VI_SUCCESS;
	ViChar wrtBuffer[BUFFER_SIZE];
    ViUInt32 retCount = 0; 
	
    if (avq7760_invalidViInt16Range (calib, 0, 2))
        return VI_ERROR_PARAMETER2;
    
    Fmt (wrtBuffer, "%s<%s", calibArr[calib]);
    
    if ((avq7760_status = viWrite (instrSession, wrtBuffer, NumFmtdBytes(), &retCount)) < 0)
        return avq7760_status;
    
    return avq7760_status;   
}

/*===========================================================================*/
/* Function: Measurement			                                         */
/* Purpose:  This function controls measurement.							 */
/*===========================================================================*/
ViStatus _VI_FUNC avq7760_Measure (ViSession instrSession, ViInt16 measure)
{
	ViStatus avq7760_status = VI_SUCCESS;
	ViChar wrtBuffer[BUFFER_SIZE];
    ViUInt32 retCount = 0; 
	
    if (avq7760_invalidViInt16Range (measure, 0, 3))
        return VI_ERROR_PARAMETER2;
    if (measure==2)
    	return VI_ERROR_PARAMETER2;
    	
    Fmt (wrtBuffer, "%s<MEA %d", measure);
    
    if ((avq7760_status = viWrite (instrSession, wrtBuffer, NumFmtdBytes(), &retCount)) < 0)
        return avq7760_status;
    
    return avq7760_status;   
}

/*===========================================================================*/
/* Function: Cursor					                                         */
/* Purpose:  This function controls cursor information window and 4 cursors. */
/*===========================================================================*/
ViStatus _VI_FUNC avq7760_Cursor (ViSession instrSession, 
								  ViBoolean domain, ViBoolean cursor,
                                  ViBoolean dispX1, ViReal64 posX1,
                                  ViBoolean dispX2, ViReal64 posX2,
                                  ViBoolean dispL1, ViReal64 posL1,
                                  ViBoolean dispL2, ViReal64 posL2)
{
	ViStatus avq7760_status = VI_SUCCESS;
	ViChar wrtBuffer[BUFFER_SIZE];
	ViReal64 rate;
	ViString unit;
    ViUInt32 retCount = 0;
    ViUInt32 total = 0;
    
	if (avq7760_invalidViBooleanRange (domain))
        return VI_ERROR_PARAMETER2;
    if (avq7760_invalidViBooleanRange (cursor))
        return VI_ERROR_PARAMETER3;
    if (avq7760_invalidViBooleanRange (dispX1))
        return VI_ERROR_PARAMETER4;
    if (avq7760_invalidViBooleanRange (dispX2))
        return VI_ERROR_PARAMETER6;
    if (avq7760_invalidViBooleanRange (dispL1))
        return VI_ERROR_PARAMETER8;
    if (avq7760_invalidViBooleanRange (dispL2))
        return VI_ERROR_INSTR_PARAMETER10;
        
    if (domain==0)     /* Wavelength */
    {
    	if (avq7760_invalidViReal64Range (posX1, 1525.0, 1635.0))
        	return VI_ERROR_PARAMETER5;
		if (avq7760_invalidViReal64Range (posX2, 1525.0, 1635.0))
        	return VI_ERROR_PARAMETER7;
    }
    
    else			   /* Frequency */
    {
    	if (avq7760_invalidViReal64Range (posX1, 183.3593, 196.5852))
        	return VI_ERROR_PARAMETER5;
		if (avq7760_invalidViReal64Range (posX2, 183.3593, 196.5852))
        	return VI_ERROR_PARAMETER7;
    }

    if (cursor==0)   /* Cursor off */					
    {
    	Fmt (wrtBuffer, "%s<CUR 0");
    	total += NumFmtdBytes();
    }
    
    else			 /* Cursor on */
    {
    	if (domain==0)
    		unit = "NM";
    	else
    		unit = "THZ";
    	
    	if (dispX1==0)
    		Fmt (wrtBuffer, "%s<XAC 0");
    	else
    		Fmt (wrtBuffer, "%s<XAS %f%s", posX1, unit);
    	total += NumFmtdBytes();
    	
    	if (dispX2==0)
    		Fmt (wrtBuffer, "%s[a]<; XBC 0");
    	else
    		Fmt (wrtBuffer, "%s[a]<; XBS %f%s", posX2, unit);
    	total += NumFmtdBytes();
    
    	if (instrStatus.displayMode==MagDisplay && instrStatus.levelScale==LogScale)
    	{
    		unit = "DB";
    		rate = 1.0;
    	}
    	else
    	{
    		unit = "N";
    		rate = 1.0e9;
    	}
    	
    	if (dispL1==0)
    		Fmt (wrtBuffer, "%s[a]<; YAC 0");
    	else
    		Fmt (wrtBuffer, "%s[a]<; YAS %f%s", posL1*rate, unit);
    	total += NumFmtdBytes();
    	
    	if (dispL2==0)
    		Fmt (wrtBuffer, "%s[a]<; YBC 0");
    	else
    		Fmt (wrtBuffer, "%s[a]<; YBS %f%s", posL2*rate, unit);
    	total += NumFmtdBytes();
    }
    
    if ((avq7760_status = viWrite (instrSession, wrtBuffer, total, &retCount)) < 0)
        return avq7760_status;
    
    return avq7760_status;   
}

/*===========================================================================*/
/* Function: Mode					                                         */
/* Purpose:  This function controls mode parameters.						 */
/*===========================================================================*/
ViStatus CVIFUNC avq7760_Mode (ViSession instrSession, 
							   ViInt16 mode, ViBoolean bandWidth, 
							   ViBoolean bwMode, ViReal64 xdBPara,
							   ViReal64 ydBPara, ViReal64 kPara)
{
	ViStatus avq7760_status = VI_SUCCESS;
	ViChar wrtBuffer[BUFFER_SIZE];
    ViUInt32 retCount = 0; 
	ViUInt32 total = 0;
	
    if (avq7760_invalidViInt16Range (mode, 0, 4))
        return VI_ERROR_PARAMETER2;
    if (mode==3)
    	return VI_ERROR_PARAMETER2;
    if (avq7760_invalidViBooleanRange (bandWidth))
        return VI_ERROR_PARAMETER3;
    if (avq7760_invalidViBooleanRange (bwMode))
        return VI_ERROR_PARAMETER4;
    if (avq7760_invalidViReal64Range (xdBPara, 0.1, 59.9))
        return VI_ERROR_PARAMETER5;
    if (avq7760_invalidViReal64Range (ydBPara, 0.1, 99.9))
        return VI_ERROR_PARAMETER6;
    if (avq7760_invalidViReal64Range (kPara, 0.1, 100.0))
        return VI_ERROR_PARAMETER7;
        
    Fmt (wrtBuffer, "%s<CUD %d", mode);
    total += NumFmtdBytes();
    
    if (bandWidth==VI_ON)
    	Fmt (wrtBuffer, "%s[a]<;SPW 1; WTY %d; WPX %f, WPY %f, WPK %f", bwMode, xdBPara, ydBPara, kPara);
    else
    	Fmt (wrtBuffer, "%s[a]<; SPW 0");
    total += NumFmtdBytes();
    
    if ((avq7760_status = viWrite (instrSession, wrtBuffer, total, &retCount)) < 0)
        return avq7760_status;
    
    return avq7760_status;   
}

/*===========================================================================*/
/* Function: Save Meas Data			                                         */
/* Purpose:  This function saves measurement results and conditions to the   */
/*			 memory or floppy.												 */
/*===========================================================================*/
ViStatus _VI_FUNC avq7760_Save_MeasData (ViSession instrSession,
                                         ViBoolean medium, ViInt16 memNum,
                                         ViString svName, ViBoolean unit,
                                         ViBoolean dtKind, ViBoolean msFrmt)
{
	ViStatus avq7760_status = VI_SUCCESS;
	ViChar wrtBuffer[BUFFER_SIZE];
    ViUInt32 retCount = 0; 
	
    if (avq7760_invalidViBooleanRange (medium))
        return VI_ERROR_PARAMETER2;
    if (medium==1)    
    	if (avq7760_invalidViInt16Range (memNum, 1, 6))
        	return VI_ERROR_PARAMETER3;
    if (avq7760_invalidViBooleanRange (unit))
        return VI_ERROR_PARAMETER5;
    if (avq7760_invalidViBooleanRange (dtKind))
        return VI_ERROR_PARAMETER6;
    if (avq7760_invalidViBooleanRange (msFrmt))
        return VI_ERROR_PARAMETER7;    
        
    if (medium==0)   /* Save to floppy */
    	Fmt (wrtBuffer, "%s<UNM %d; SAV#%s#", unit, svName);
    else			 /* Save to memory */
    	if (dtKind==0)   /* Data kind is dispalyed data */
    		Fmt (wrtBuffer, "%s<UNM %d; SMS 0; SAV%d#%s#", unit, memNum, svName);
    	else			 /* Data kind is measured data */
    		Fmt (wrtBuffer, "%s<UNM %d; SMS 1; ASC %d; SAV%d#%s#", unit, msFrmt, memNum, svName);
    	
    if ((avq7760_status = viWrite (instrSession, wrtBuffer, NumFmtdBytes(), &retCount)) < 0)
        return avq7760_status;
    
    return avq7760_status;
}

/*===========================================================================*/
/* Function: Save Panel				                                         */
/* Purpose:  This function saves measurement conditions to the memory or 	 */
/*			 floppy.												 		 */
/*===========================================================================*/
ViStatus _VI_FUNC avq7760_Save_Panel (ViSession instrSession,
                                      ViBoolean medium, ViInt16 memNum,
                                      ViString svName)
{
	ViStatus avq7760_status = VI_SUCCESS;
	ViChar wrtBuffer[BUFFER_SIZE];
    ViUInt32 retCount = 0; 
	
    if (avq7760_invalidViBooleanRange (medium))
        return VI_ERROR_PARAMETER2;
    if (medium==1)
    	if (avq7760_invalidViInt16Range (memNum, 1, 10))
        	return VI_ERROR_PARAMETER3;
        
    if (medium==0)   /* Save to floppy */
    	Fmt (wrtBuffer, "%s<SVP#%s#", svName);
    else			 /* Save to memory */
    	Fmt (wrtBuffer, "%s<SVP%d#%s#", memNum, svName);
    	
    if ((avq7760_status = viWrite (instrSession, wrtBuffer, NumFmtdBytes(), &retCount)) < 0)
        return avq7760_status;
    
    return avq7760_status;
}

/*===========================================================================*/
/* Function: Save Bitmap			                                         */
/* Purpose:  This control saves displayed data as bitmap format to the		 */
/*			 floppy disk.													 */
/*===========================================================================*/
ViStatus _VI_FUNC avq7760_Save_Bitmap (ViSession instrSession)
{
	ViStatus avq7760_status = VI_SUCCESS;
    ViUInt32 retCount = 0; 

    if ((avq7760_status = viWrite (instrSession, "BIT", 3, &retCount)) < 0)
        return avq7760_status;
    
    return avq7760_status;
    
}

/*===========================================================================*/
/* Function: Load Meas Data			                                         */
/* Purpose:  This function loads measurement results from memory or floppy.	 */
/*			 When loading from memory, only displayed data is loaded.		 */
/*			 When loading from floppy and data are saved as MEA-BIN mode,    */
/*			 all measured data are loaded. Otherwise, only displayed data is */
/*			 loaded.														 */
/*===========================================================================*/
ViStatus _VI_FUNC avq7760_Load_MeasData (ViSession instrSession,
                                         ViBoolean medium, ViInt16 memNum,
                                         ViString flName)
{
	ViStatus avq7760_status = VI_SUCCESS;
	ViChar wrtBuffer[BUFFER_SIZE];
    ViUInt32 retCount = 0; 
	
    if (avq7760_invalidViBooleanRange (medium))
        return VI_ERROR_PARAMETER2;
    if (medium==1)    
    	if (avq7760_invalidViInt16Range (memNum, 1, 6))
        	return VI_ERROR_PARAMETER3;
       
    if (medium==0)   /* Load from floppy */
    	Fmt (wrtBuffer, "%s<RCL #%s#", flName);
    else			 /* Load from memory */
    	Fmt (wrtBuffer, "%s<RCL %d", memNum);
    		
    if ((avq7760_status = viWrite (instrSession, wrtBuffer, NumFmtdBytes(), &retCount)) < 0)
        return avq7760_status;
    
    return avq7760_status;
}

/*===========================================================================*/
/* Function: Load Panel				                                         */
/* Purpose:  This function loads measurement conditions from memory or 		 */
/*			 floppy.	 													 */
/*===========================================================================*/
ViStatus _VI_FUNC avq7760_Load_Panel (ViSession instrSession,
                                      ViBoolean medium, ViInt16 memNum,
                                      ViString flName)
{
	ViStatus avq7760_status = VI_SUCCESS;
	ViChar wrtBuffer[BUFFER_SIZE];
    ViUInt32 retCount = 0; 
	
    if (avq7760_invalidViBooleanRange (medium))
        return VI_ERROR_PARAMETER2;
    if (medium==1)    
    	if (avq7760_invalidViInt16Range (memNum, 1, 10))
        	return VI_ERROR_PARAMETER3;
       
    if (medium==0)   /* Load from floppy */
    	Fmt (wrtBuffer, "%s<RCP #%s#", flName);
    else			 /* Load from memory */
    	Fmt (wrtBuffer, "%s<RCP %d", memNum);
    		
    if ((avq7760_status = viWrite (instrSession, wrtBuffer, NumFmtdBytes(), &retCount)) < 0)
        return avq7760_status;
    
    return avq7760_status;
}

/*===========================================================================*/
/* Function: Color Pattern                                                   */
/* Purpose:  This function sets the color pattern of the instrument's 		 */
/*			 screen.														 */
/*===========================================================================*/
ViStatus CVIFUNC avq7760_ColorPattern (ViSession instrSession, ViInt16 clrPttrn)
{
	ViStatus avq7760_status = VI_SUCCESS;
	ViChar wrtBuffer[BUFFER_SIZE];
    ViUInt32 retCount = 0; 
	
    if (avq7760_invalidViInt16Range (clrPttrn, 0, 4))
        return VI_ERROR_PARAMETER2;
    
    Fmt (wrtBuffer, "%s<CPT %d", clrPttrn);
    
    if ((avq7760_status = viWrite (instrSession, wrtBuffer, NumFmtdBytes(), &retCount)) < 0)
        return avq7760_status;
    
    return avq7760_status;
}

/*===========================================================================*/
/* Function: Printer	                                                     */
/* Purpose:  This function sets printer type and switches to execute.		 */
/*===========================================================================*/
ViStatus CVIFUNC avq7760_Printer (ViSession instrSession,
                                  ViInt16 prtType, ViBoolean exe)
{
	ViStatus avq7760_status = VI_SUCCESS;
	ViChar wrtBuffer[BUFFER_SIZE];
    ViUInt32 retCount = 0; 
	ViUInt32 total = 0;
	
    if (avq7760_invalidViInt16Range (prtType, 0, 2))
        return VI_ERROR_PARAMETER2;
    if (avq7760_invalidViBooleanRange (exe))
        return VI_ERROR_PARAMETER3; 
        
    Fmt (wrtBuffer, "%s<PRT%d", prtType);
    total += NumFmtdBytes();
    
    if (exe==VI_ON)
    {
    	Fmt (wrtBuffer, "%s[a]<; EPR");
    	total += NumFmtdBytes();
    }
    
    if ((avq7760_status = viWrite (instrSession, wrtBuffer, total, &retCount)) < 0)
        return avq7760_status;
    
    return avq7760_status;
}

/*===========================================================================*/
/* Function: Label	                                                     	 */
/* Purpose:  This function writes label to the Q7760.						 */
/*===========================================================================*/
ViStatus CVIFUNC avq7760_Label (ViSession instrSession, ViString label)
{
    ViUInt32 retCnt = 0;
    ViChar wrtBuf[53];
    ViStatus avq7760_status = VI_SUCCESS;

    Fmt (wrtBuf, "%s<LAB#%s#",label);    
    
    if ((avq7760_status = viWrite (instrSession, wrtBuf, NumFmtdBytes(), &retCnt)) < 0)
      	return avq7760_status;

    return avq7760_status;     
}

/*===========================================================================*/
/* Function: Buzzer                                                     	 */
/* Purpose:  This function sets the conditions for sounding the buzzer.		 */
/*===========================================================================*/
ViStatus CVIFUNC avq7760_Buzzer (ViSession instrSession, 
								 ViBoolean beep, ViBoolean warn, 
								 ViBoolean quiet)
{
	ViStatus avq7760_status = VI_SUCCESS;
	ViChar wrtBuffer[BUFFER_SIZE];
    ViUInt32 retCount = 0; 

    if (avq7760_invalidViBooleanRange (beep))
        return VI_ERROR_PARAMETER2;
    if (avq7760_invalidViBooleanRange (warn))
        return VI_ERROR_PARAMETER3;
    if (avq7760_invalidViBooleanRange (quiet))
        return VI_ERROR_PARAMETER4;    
    
    Fmt (wrtBuffer, "%s<BUZ %d; WAR %d; QUI %d", beep, warn, quiet);
    
    if ((avq7760_status = viWrite (instrSession, wrtBuffer, NumFmtdBytes(), &retCount)) < 0)
        return avq7760_status;
    
    return avq7760_status;
}

/*===========================================================================*/
/* Function: Set Clock                                                       */
/* Purpose:  This function sets the instrument's clock.						 */
/*===========================================================================*/
ViStatus CVIFUNC avq7760_Set_Clock (ViSession instrSession, 
									ViInt16 year, ViInt16 month, 
									ViInt16 day, ViInt16 hour,
                                    ViInt16 minute)
{
	ViStatus avq7760_status = VI_SUCCESS;
	ViChar wrtBuffer[BUFFER_SIZE];
    ViUInt32 retCount = 0; 

    if (avq7760_invalidViInt16Range (year, 0, 99))
        return VI_ERROR_PARAMETER2;
    if (avq7760_invalidViInt16Range (month, 1, 12))
        return VI_ERROR_PARAMETER3;
    if (avq7760_invalidViInt16Range (day, 1, 31))
        return VI_ERROR_PARAMETER4;
    if (avq7760_invalidViInt16Range (hour, 0, 23))
        return VI_ERROR_PARAMETER5;
    if (avq7760_invalidViInt16Range (minute, 0, 59))
        return VI_ERROR_PARAMETER6;
        
    Fmt (wrtBuffer, "%s<CLO #%d[w2p0]-%d[w2p0]-%d[w2p0],%d[w2p0]:%d[w2p0]:00#", year, month, day, hour, minute);
    
    if ((avq7760_status = viWrite (instrSession, wrtBuffer, NumFmtdBytes(), &retCount)) < 0)
        return avq7760_status;
    
    return avq7760_status; 
}

/*===========================================================================*/
/* Function: Display Clock                                                   */
/* Purpose:  This function switches the clock display on the instrument's	 */
/*			 screen.														 */
/*===========================================================================*/
ViStatus CVIFUNC avq7760_Disp_Clock (ViSession instrSession, ViBoolean display)
{
	ViStatus avq7760_status = VI_SUCCESS;
	ViChar wrtBuffer[BUFFER_SIZE];
    ViUInt32 retCount = 0; 

    if (avq7760_invalidViBooleanRange (display))
        return VI_ERROR_PARAMETER2;
        
    Fmt (wrtBuffer, "%s<CKD %d", display);
    
    if ((avq7760_status = viWrite (instrSession, wrtBuffer, NumFmtdBytes(), &retCount)) < 0)
        return avq7760_status;
    
    return avq7760_status; 
}

/*===========================================================================*/
/* Function: SRQ Control                                                     */
/* Purpose:  This function switches SRQ signal transmission.				 */
/*===========================================================================*/
ViStatus CVIFUNC avq7760_SRQ (ViSession instrSession, ViBoolean outSRQ)
{
	ViStatus avq7760_status = VI_SUCCESS;
	ViChar wrtBuffer[BUFFER_SIZE];
    ViUInt32 retCount = 0; 

    if (avq7760_invalidViBooleanRange (outSRQ))
        return VI_ERROR_PARAMETER2;
    
    Fmt (wrtBuffer, "%s<SRQ %d", outSRQ);
    
    if ((avq7760_status = viWrite (instrSession, wrtBuffer, NumFmtdBytes(), &retCount)) < 0)
        return avq7760_status;
    
    return avq7760_status; 
}

/*===========================================================================*/
/* Function: Mask Status Byte                                                */
/* Purpose:  This function masks the status byte.							 */
/*===========================================================================*/
ViStatus CVIFUNC avq7760_Mask_STB (ViSession instrSession, ViInt16 mask)
{
	ViStatus avq7760_status = VI_SUCCESS;
	ViChar wrtBuffer[BUFFER_SIZE];
    ViUInt32 retCount = 0; 

    if (avq7760_invalidViInt16Range (mask, 0, 255))
        return VI_ERROR_PARAMETER2;
    
    Fmt (wrtBuffer, "%s<MSK %d", mask);
    
    if ((avq7760_status = viWrite (instrSession, wrtBuffer, NumFmtdBytes(), &retCount)) < 0)
        return avq7760_status;
    
    return avq7760_status;     
}

/*===========================================================================*/
/* Function: Clear Status Byte                                               */
/* Purpose:  This function Clears the status byte.							 */
/*===========================================================================*/
ViStatus CVIFUNC avq7760_Clear_STB (ViSession instrSession)
{
	ViStatus avq7760_status = VI_SUCCESS;
	ViUInt32 retCount = 0;    
	
	if ((avq7760_status = viWrite (instrSession, "CSB", 3, &retCount)) < 0)
        return avq7760_status;
    
    return avq7760_status;  
}

/*===========================================================================*/
/* Function: Header				                                             */
/* Purpose:  This function switches to add or not the header, when 			 */
/*			 outputting GPIB data.											 */
/*===========================================================================*/
ViStatus CVIFUNC avq7760_Header (ViSession instrSession, ViBoolean header)
{
	ViStatus avq7760_status = VI_SUCCESS;
	ViChar wrtBuffer[BUFFER_SIZE];
    ViUInt32 retCount = 0; 

    if (avq7760_invalidViBooleanRange (header))
        return VI_ERROR_PARAMETER2;
    
    Fmt (wrtBuffer, "%s<HED %d", header);
    
    if ((avq7760_status = viWrite (instrSession, wrtBuffer, NumFmtdBytes(), &retCount)) < 0)
        return avq7760_status;
    
    return avq7760_status; 
}

/*===========================================================================*/
/* Function: Set Terminator		                                             */
/* Purpose:  This function sets the terminator.								 */
/*===========================================================================*/
ViStatus CVIFUNC avq7760_Set_Terminator (ViSession instrSession, 
										 ViInt16 terminator)
{
	ViStatus avq7760_status = VI_SUCCESS;
	ViChar wrtBuffer[BUFFER_SIZE];
    ViUInt32 retCount = 0; 

    if (avq7760_invalidViInt16Range (terminator, 0, 3))
        return VI_ERROR_PARAMETER2;
    
    Fmt (wrtBuffer, "%s<DEL %d", terminator);
    
    if ((avq7760_status = viWrite (instrSession, wrtBuffer, NumFmtdBytes(), &retCount)) < 0)
        return avq7760_status;
    
    return avq7760_status; 
}

/*===========================================================================*/
/* Function: Set Data Separator                                              */
/* Purpose:  This function selects the data separator which is used when	 */
/*			 outputting GPIB data.											 */
/*===========================================================================*/
ViStatus CVIFUNC avq7760_Set_DataSepa (ViSession instrSession, ViInt16 dataSepa)
{
	ViStatus avq7760_status = VI_SUCCESS;
	ViChar wrtBuffer[BUFFER_SIZE];
    ViUInt32 retCount = 0; 

    if (avq7760_invalidViInt16Range (dataSepa, 0, 3))
        return VI_ERROR_PARAMETER2;
    
    Fmt (wrtBuffer, "%s<SDL %d", dataSepa);
    
    if ((avq7760_status = viWrite (instrSession, wrtBuffer, NumFmtdBytes(), &retCount)) < 0)
        return avq7760_status;
    
    return avq7760_status; 
}

/*===========================================================================*/
/* Function: Set Data Output parameters                                      */
/* Purpose:  This function sets the parameters for the GPIB data output.	 */
/*===========================================================================*/
ViStatus CVIFUNC avq7760_Sel_DataOut (ViSession instrSession,
                                      ViInt16 format, ViBoolean screen)
{
	ViStatus avq7760_status = VI_SUCCESS;
	ViChar wrtBuffer[BUFFER_SIZE];
    ViUInt32 retCount = 0; 

    if (avq7760_invalidViInt16Range (format, 0, 3))
        return VI_ERROR_PARAMETER2;
    if (avq7760_invalidViBooleanRange (screen))
        return VI_ERROR_PARAMETER3;    
    
    Fmt (wrtBuffer, "%s<FMT %d; OVS %d", format, screen);
    
    if ((avq7760_status = viWrite (instrSession, wrtBuffer, NumFmtdBytes(), &retCount)) < 0)
        return avq7760_status;
    
    return avq7760_status; 
}

/*===========================================================================*/
/* Function: Read Waveform Data by ASCII Mode                                */
/* Purpose:  This function returns waveform data by ASCII Mode.				 */
/*===========================================================================*/
ViStatus CVIFUNC avq7760_ASCII_Waveform (ViSession instrSession,
                                         ViBoolean screen, ViBoolean axis,
                                         ViInt16 *dtNum, ViReal64 waveform[])
{
	ViStatus avq7760_status = VI_SUCCESS;
	ViChar *buffer;
	ViChar fmtBuffer[20];
    ViUInt32 retCount = 0; 

    if (avq7760_invalidViBooleanRange (screen))
        return VI_ERROR_PARAMETER2;
    if (avq7760_invalidViBooleanRange (axis))
        return VI_ERROR_PARAMETER3;

    if ((avq7760_status = avq7760_Sel_DataOut (instrSession, 0, screen)) < 0)
        return avq7760_status;
    
    buffer = malloc(252042);    
    setCntrl (instrSession);
    
    if ((avq7760_status = viWrite (instrSession, "ODN?", 4, &retCount)) < 0)                
	    return avq7760_status;                                                            
	if ((avq7760_status = viRead (instrSession, buffer, BUFFER_SIZE, &retCount)) < 0)        
	    return avq7760_status;                                                            
	if (Scan (buffer, "%s>%d[b2]", dtNum) < 1)               
	    return VI_ERROR_INSTR_INTERPRETING_RESPONSE;
	    
	Fmt (buffer, "%s<OSD %d", axis);
	if ((avq7760_status = viWrite (instrSession, buffer, NumFmtdBytes(), &retCount)) < 0)                
	    return avq7760_status;                                                            
	if ((avq7760_status = viRead (instrSession, buffer, *dtNum * 21, &retCount)) < 0)        
	    return avq7760_status;
	Fmt (fmtBuffer, "%s<%s%d%s", "%s>%", *dtNum, "f[x]");
	if (Scan (buffer, fmtBuffer, waveform) < 1)               
	    return VI_ERROR_INSTR_INTERPRETING_RESPONSE;
    
    free (buffer);
    resetCntrl (instrSession);
    
	return avq7760_status; 
}

/*===========================================================================*/
/* Function: Read Waveform Data by Binary 16 bit Mode                        */
/* Purpose:  This function returns waveform data by binary 16 bit mode.		 */
/*===========================================================================*/
void bin_ViInt16 (int size, char *src, ViInt16 *dst)
{
	unsigned char *pt_src;
	
	pt_src = (unsigned char *)src;
	
	while (--size>=0)
	{
		*dst    = *pt_src++ * 256;
		*dst++ += *pt_src++;
	}
}

ViStatus CVIFUNC avq7760_BIN16_Waveform (ViSession instrSession,
                                         ViBoolean screen, ViBoolean axis,
                                         ViInt16 *dtNum, ViInt16 waveform[])
{
	ViStatus avq7760_status = VI_SUCCESS;
	ViChar buffer[24004];
    ViUInt32 retCount = 0;
    ViBoolean termRead;

    if (avq7760_invalidViBooleanRange (screen))
        return VI_ERROR_PARAMETER2;
    if (avq7760_invalidViBooleanRange (axis))
        return VI_ERROR_PARAMETER3;
    
    if ((avq7760_status = avq7760_Sel_DataOut (instrSession, 1, screen)) < 0)
        return avq7760_status;
    
    setCntrl (instrSession);
    
    if ((avq7760_status = viWrite (instrSession, "ODN?", 4, &retCount)) < 0)                
	    return avq7760_status;                                                            
	if ((avq7760_status = viRead (instrSession, buffer, BUFFER_SIZE, &retCount)) < 0)        
	    return avq7760_status;                                                            
	if (Scan (buffer, "%s>%d[b2]", dtNum) < 1)               
	    return VI_ERROR_INSTR_INTERPRETING_RESPONSE;
	
	if ((avq7760_status = viGetAttribute (instrSession, VI_ATTR_TERMCHAR_EN, &termRead)) < 0)
		return avq7760_status;
	if ((avq7760_status = viSetAttribute (instrSession, VI_ATTR_TERMCHAR_EN, VI_FALSE)) < 0)
		return avq7760_status;
				  
	Fmt (buffer, "%s<OSD %d", axis);
	if ((avq7760_status = viWrite (instrSession, buffer, NumFmtdBytes(), &retCount)) < 0)                
	    return avq7760_status;                                                            
	if ((avq7760_status = viRead (instrSession, buffer, *dtNum * 2, &retCount)) < 0)        
	    return avq7760_status;
	
	if ((avq7760_status = viSetAttribute (instrSession, VI_ATTR_TERMCHAR_EN, termRead)) < 0)
		return avq7760_status;
	
	bin_ViInt16 (*dtNum, buffer, waveform);
	    
    resetCntrl (instrSession);
    
	return avq7760_status; 
}

/*===========================================================================*/
/* Function: Read Waveform Data by Binary 64 bit Float Mode                  */
/* Purpose:  This function returns waveform data by binary 64 bit float 	 */
/*			 mode.		 													 */
/*===========================================================================*/
void bin_ViReal64 (int size, char *src, ViReal64 *dst)
{
	unsigned char buf[8];
	unsigned char *pt_src, *pt_dst;
	ViInt16       n;
	ViReal64      *dbl_src, *pt_buf;
	
	pt_buf  = (ViReal64 *)buf;
	dbl_src = (ViReal64 *)src;
	
	while (--size>=0)
	{
		pt_src = (unsigned char *)dbl_src + 7;
		pt_dst = buf;
		
		n = 8;
		while (--n>0)
			*pt_dst++ = *pt_src--;
		*pt_dst = *pt_src;	
			
		*dst++ = *pt_buf;
		dbl_src++;
	}
}

ViStatus CVIFUNC avq7760_BIN64Float_Waveform (ViSession instrSession,
                                              ViBoolean screen, ViBoolean axis,
                                              ViInt16 *dtNum, ViReal64 waveform[])
{
	ViStatus avq7760_status = VI_SUCCESS;
	ViChar buffer[96016];
    ViUInt32 retCount = 0;
    ViBoolean termRead;
 
    if (avq7760_invalidViBooleanRange (screen))
        return VI_ERROR_PARAMETER2;
    if (avq7760_invalidViBooleanRange (axis))
        return VI_ERROR_PARAMETER3;
    
    if ((avq7760_status = avq7760_Sel_DataOut (instrSession, 2, screen)) < 0)
        return avq7760_status;

    setCntrl (instrSession);
    
    if ((avq7760_status = viWrite (instrSession, "ODN?", 4, &retCount)) < 0)                
	    return avq7760_status;                                                            
	if ((avq7760_status = viRead (instrSession, buffer, BUFFER_SIZE, &retCount)) < 0)        
	    return avq7760_status;                                                            
	if (Scan (buffer, "%s>%d[b2]", dtNum) < 1)               
	    return VI_ERROR_INSTR_INTERPRETING_RESPONSE;
	
	if ((avq7760_status = viGetAttribute (instrSession, VI_ATTR_TERMCHAR_EN, &termRead)) < 0)
		return avq7760_status;
	if ((avq7760_status = viSetAttribute (instrSession, VI_ATTR_TERMCHAR_EN, VI_FALSE)) < 0)
		return avq7760_status;
	
	Fmt (buffer, "%s<OSD %d", axis);
	if ((avq7760_status = viWrite (instrSession, buffer, NumFmtdBytes(), &retCount)) < 0)                
	    return avq7760_status;                                                            
	if ((avq7760_status = viRead (instrSession, buffer, *dtNum * 8, &retCount)) < 0)        
	    return avq7760_status;
	
	if ((avq7760_status = viSetAttribute (instrSession, VI_ATTR_TERMCHAR_EN, termRead)) < 0)
		return avq7760_status;
	
	bin_ViReal64 (*dtNum, buffer, waveform);
    
    resetCntrl (instrSession);
    
	return avq7760_status; 
}

/*===========================================================================*/
/* Function: Read Waveform Data by Binary 32 bit Float Mode                  */
/* Purpose:  This function returns waveform data by binary 32 bit float 	 */
/*			 mode.		 													 */
/*===========================================================================*/
void bin_float (int size, char *src, float *dst)
{
	unsigned char buf[4];
	unsigned char *pt_src, *pt_dst;
	ViInt16       n;
	float         *flt_src, *pt_buf;
	
	pt_buf  = (float *)buf;
	flt_src = (float *)src;
	
	while (--size>=0)
	{
		pt_src = (unsigned char *)flt_src + 3;
		pt_dst = buf;
		
		n = 4;
		while (--n>0)
			*pt_dst++ = *pt_src--;
		*pt_dst = *pt_src;	
			
		*dst++ = *pt_buf;
		flt_src++;
	}
}

ViStatus CVIFUNC avq7760_BIN32Float_Waveform (ViSession instrSession,
                                              ViBoolean screen, ViBoolean axis,
                                              ViInt16 *dtNum, float waveform[])
{
	ViStatus avq7760_status = VI_SUCCESS;
	ViChar buffer[48008];
    ViUInt32 retCount = 0;
    ViBoolean termRead;
    
    if (avq7760_invalidViBooleanRange (screen))
        return VI_ERROR_PARAMETER2;
    if (avq7760_invalidViBooleanRange (axis))
        return VI_ERROR_PARAMETER3;
    
    if ((avq7760_status = avq7760_Sel_DataOut (instrSession, 3, screen)) < 0)
        return avq7760_status;
    
    setCntrl (instrSession);
    
    if ((avq7760_status = viWrite (instrSession, "ODN?", 4, &retCount)) < 0)                
	    return avq7760_status;                                                            
	if ((avq7760_status = viRead (instrSession, buffer, BUFFER_SIZE, &retCount)) < 0)        
	    return avq7760_status;                                                            
	if (Scan (buffer, "%s>%d[b2]", dtNum) < 1)               
	    return VI_ERROR_INSTR_INTERPRETING_RESPONSE;
	
	if ((avq7760_status = viGetAttribute (instrSession, VI_ATTR_TERMCHAR_EN, &termRead)) < 0)
		return avq7760_status;
	if ((avq7760_status = viSetAttribute (instrSession, VI_ATTR_TERMCHAR_EN, VI_FALSE)) < 0)
		return avq7760_status;
				  
	Fmt (buffer, "%s<OSD %d", axis);
	if ((avq7760_status = viWrite (instrSession, buffer, NumFmtdBytes(), &retCount)) < 0)                
	    return avq7760_status;                                                            
	if ((avq7760_status = viRead (instrSession, buffer, *dtNum * 4, &retCount)) < 0)        
	    return avq7760_status;

	if ((avq7760_status = viSetAttribute (instrSession, VI_ATTR_TERMCHAR_EN, termRead)) < 0)
		return avq7760_status;
	
	bin_float (*dtNum, buffer, waveform);
    
    resetCntrl (instrSession);
    
	return avq7760_status; 
}

/*===========================================================================*/
/* Function: Fiber Length Data									             */
/* Purpose:  This function returns the fiber length data.					 */
/*===========================================================================*/
ViStatus CVIFUNC avq7760_FiberLengthData (ViSession instrSession, 
										  ViReal64 *fiberLength)
{
	ViStatus avq7760_status = VI_SUCCESS;
	ViChar rdBuffer[BUFFER_SIZE];
    ViUInt32 retCount = 0;
    
    setCntrl (instrSession);
    
    if ((avq7760_status = viWrite (instrSession, "OFL?", 4, &retCount)) < 0)                
	    return avq7760_status;                                                            
	if ((avq7760_status = viRead (instrSession, rdBuffer, BUFFER_SIZE, &retCount)) < 0)        
	    return avq7760_status;                                                            
	if (Scan (rdBuffer, "%s>%f", fiberLength) < 1)               
	    return VI_ERROR_INSTR_INTERPRETING_RESPONSE;
	    
    resetCntrl (instrSession);
    
	return avq7760_status; 
}

/*===========================================================================*/
/* Function: Peak Search Data									             */
/* Purpose:  This function returns the peak search data.					 */
/*===========================================================================*/
ViStatus CVIFUNC avq7760_PeakSearchData (ViSession instrSession,
                                         ViReal64 *pos, ViReal64 *pkLvl)
{
	ViStatus avq7760_status = VI_SUCCESS;
	ViChar rdBuffer[BUFFER_SIZE];
    ViUInt32 retCount = 0;
    
    setCntrl (instrSession);
    
    if ((avq7760_status = viWrite (instrSession, "OPK?", 4, &retCount)) < 0)                
	    return avq7760_status;                                                            
	if ((avq7760_status = viRead (instrSession, rdBuffer, BUFFER_SIZE, &retCount)) < 0)        
	    return avq7760_status;                                                            
	if (Scan (rdBuffer, "%s>%f,%f", pos, pkLvl) < 2)               
	    return VI_ERROR_INSTR_INTERPRETING_RESPONSE;
	    
    resetCntrl (instrSession);
    
	return avq7760_status; 
}

/*===========================================================================*/
/* Function: Cursor Data										             */
/* Purpose:  This function returns the cursor data.							 */
/*===========================================================================*/
ViStatus CVIFUNC avq7760_CursorData (ViSession instrSession,
                                     ViInt16 csrMode, ViReal64 *data1,
                                     ViReal64 *data2, ViReal64 *data3,
                                     ViReal64 *data4, ViReal64 *data5,
                                     ViReal64 *data6)
{
	ViStatus avq7760_status = VI_SUCCESS;
	ViChar buffer[BUFFER_SIZE];
    ViUInt32 retCount = 0;
    
    if (avq7760_invalidViInt16Range (csrMode, 0, 3))
        return VI_ERROR_PARAMETER2;
        
    setCntrl (instrSession);
    
    Fmt (buffer, "CUD %d", csrMode);
    if ((avq7760_status = viWrite (instrSession, buffer, NumFmtdBytes(), &retCount)) < 0)                
	    return avq7760_status;
	    
    if ((avq7760_status = viWrite (instrSession, "OCD?", 4, &retCount)) < 0)                
	    return avq7760_status;                                                            
	if ((avq7760_status = viRead (instrSession, buffer, BUFFER_SIZE, &retCount)) < 0)        
	    return avq7760_status;
	    
	if (csrMode<=1)
	{
		if (Scan (buffer, "%s>%f,%f,%f,%f,%f,%f", data1, data2, data3, data4, data5, data6) < 6)               
	    	return VI_ERROR_INSTR_INTERPRETING_RESPONSE;
	}
	else
	{
		if (Scan (buffer, "%s>%f,%f,%f,%f", data1, data2, data3, data4) < 4)               
	    	return VI_ERROR_INSTR_INTERPRETING_RESPONSE;
	    *data5 = 0.0;
	    *data6 = 0.0;
	}
    
    resetCntrl (instrSession);
    
    return avq7760_status;
    
}

/*===========================================================================*/
/* Function: Half Width Data									             */
/* Purpose:  This function returns the calculated half width data.			 */
/*===========================================================================*/
ViStatus CVIFUNC avq7760_HalfWidthData (ViSession instrSession,
                                        ViReal64 *center, ViReal64 *hfWidth,
                                        ViInt16 *pkNum)
{
	ViStatus avq7760_status = VI_SUCCESS;
	ViChar rdBuffer[BUFFER_SIZE];
    ViUInt32 retCount = 0;
    
    setCntrl (instrSession);
    
    if ((avq7760_status = viWrite (instrSession, "OSW?", 4, &retCount)) < 0)                
	    return avq7760_status;                                                            
	if ((avq7760_status = viRead (instrSession, rdBuffer, BUFFER_SIZE, &retCount)) < 0)        
	    return avq7760_status;                                                            
	if (Scan (rdBuffer, "%s>%f,%f,%d[b2]", center, hfWidth, pkNum) < 3)               
	    return VI_ERROR_INSTR_INTERPRETING_RESPONSE;
    
    resetCntrl (instrSession);
    
	return avq7760_status; 
}

/*===========================================================================*/
/* Function: Clock Data                                                      */
/* Purpose:  This function returns the time and date which are setted at the */
/*			 instrument's clock.  											 */
/*===========================================================================*/
ViStatus CVIFUNC avq7760_ClockData (ViSession instrSession, 
									ViInt16 *year, ViInt16 *month, 
									ViInt16 *day, ViInt16 *hour,
                                    ViInt16 *minute, ViInt16 *second)
{
	ViStatus avq7760_status = VI_SUCCESS;
	ViChar rdBuffer[BUFFER_SIZE];
    ViUInt32 retCount = 0;
    
    setCntrl (instrSession);
    
    if ((avq7760_status = viWrite (instrSession, "CLO?", 4, &retCount)) < 0)                
	    return avq7760_status;                                                            
	if ((avq7760_status = viRead (instrSession, rdBuffer, BUFFER_SIZE, &retCount)) < 0)        
	    return avq7760_status;                                                            
	if (Scan (rdBuffer, "%s>%d[b2]-%d[b2]-%d[b2],%d[b2]:%d[b2]:%d[b2]", year, month, day, hour, minute, second) < 6)               
	    return VI_ERROR_INSTR_INTERPRETING_RESPONSE;
    
    resetCntrl (instrSession);
    
	return avq7760_status; 
}

/*===========================================================================*/
/* Function: Reset                                                           */
/* Purpose:  This function resets the instrument to the state when power is  */
/*			 set to on.														 */
/*===========================================================================*/
ViStatus _VI_FUNC avq7760_Reset (ViSession instrSession)
{
    ViStatus avq7760_status = VI_SUCCESS;
	ViUInt32 retCount = 0;  

    /*  Initialize the instrument to the state when power is set to on.  */
    if ((avq7760_status = viWrite (instrSession, "C", 1, &retCount)) < 0)
        return avq7760_status;

    return avq7760_status;
}

/*===========================================================================*/
/* Function: Reset                                                           */
/* Purpose:  This function resets the instrument to the pre-defined state. 	 */
/*===========================================================================*/
ViStatus CVIFUNC avq7760_Preset (ViSession instrSession)
{
    ViStatus avq7760_status = VI_SUCCESS;
	ViUInt32 retCount = 0;  

    /*  Initialize the instrument to the state when power is set to on.  */
    if ((avq7760_status = viWrite (instrSession, "IPR", 3, &retCount)) < 0)
        return avq7760_status;
        
    return avq7760_status;
}

/*===========================================================================*/
/* Function: Self-Test                                                       */
/* Purpose:  This function causes the instrument to perform a self-test and  */
/*			 returns the results.											 */
/*===========================================================================*/
ViStatus _VI_FUNC avq7760_SelfTest (ViSession instrSession, ViInt16 *testResult)
{
	ViStatus avq7760_status = VI_SUCCESS;
	ViChar rdBuffer[BUFFER_SIZE];
    ViUInt32 retCount = 0;
    ViUInt32 timeOut;
    
	if ((viGetAttribute (instrSession, VI_ATTR_TMO_VALUE, &timeOut)) < 0)
		avq7760_status;
	if ((viSetAttribute (instrSession, VI_ATTR_TMO_VALUE, 300000)) < 0)
		avq7760_status;
	
    if ((avq7760_status = viWrite (instrSession, "*TST?", 5, &retCount)) < 0)                
	    return avq7760_status;                                                            
	if ((avq7760_status = viRead (instrSession, rdBuffer, BUFFER_SIZE, &retCount)) < 0)        
	    return avq7760_status;                                                            
	if (Scan (rdBuffer, "%s>%d[b2]", testResult) < 1)                 
	    return VI_ERROR_INSTR_INTERPRETING_RESPONSE;
	
	if ((viSetAttribute (instrSession, VI_ATTR_TMO_VALUE, timeOut)) < 0)
		avq7760_status;
	
	return avq7760_status; 
}

/*===========================================================================*/
/* Function: Error Message                                                   */
/* Purpose:  This function translates the error return value from the        */
/*           instrument driver into a user-readable string.                  */
/*===========================================================================*/
ViStatus _VI_FUNC avq7760_ErrorMessage (ViSession instrSession,
                                         ViStatus statusCode, ViChar errMessage[])
{
    ViStatus avq7760_status = VI_SUCCESS;
    ViInt16 i;
    static avq7760_tStringValPair statusDescArray[] = {
        {VI_WARN_NSUP_ID_QUERY,     "WARNING: ID Query not supported"},
        {VI_WARN_NSUP_RESET,        "WARNING: Reset not supported"},
        {VI_WARN_NSUP_SELF_TEST,    "WARNING: Self-test not supported"},
        {VI_WARN_NSUP_ERROR_QUERY,  "WARNING: Error Query not supported"},     
        {VI_WARN_NSUP_REV_QUERY,    "WARNING: Revision Query not supported"},
        {VI_ERROR_PARAMETER1,   "ERROR: Parameter 1 out of range"},
        {VI_ERROR_PARAMETER2,   "ERROR: Parameter 2 out of range"},
        {VI_ERROR_PARAMETER3,   "ERROR: Parameter 3 out of range"},
        {VI_ERROR_PARAMETER4,   "ERROR: Parameter 4 out of range"},
        {VI_ERROR_PARAMETER5,   "ERROR: Parameter 5 out of range"},
        {VI_ERROR_PARAMETER6,   "ERROR: Parameter 6 out of range"},
        {VI_ERROR_PARAMETER7,   "ERROR: Parameter 7 out of range"},
        {VI_ERROR_PARAMETER8,   "ERROR: Parameter 8 out of range"},
        {VI_ERROR_FAIL_ID_QUERY,"ERROR: Identification query failed"},
        {VI_ERROR_INV_RESPONSE, "ERROR: Interpreting instrument response"},
        {VI_ERROR_INSTR_FILE_OPEN,    "ERROR: Opening the specified file"},
        {VI_ERROR_INSTR_FILE_WRITE,   "ERROR: Writing to the specified file"},
        {VI_ERROR_INSTR_INTERPRETING_RESPONSE, "ERROR: Interpreting the instrument's response"},
        {VI_ERROR_INSTR_PARAMETER9 ,  "ERROR: Parameter 9 out of range"},
        {VI_ERROR_INSTR_PARAMETER10,  "ERROR: Parameter 10 out of range"},
        {VI_ERROR_INSTR_PARAMETER11,  "ERROR: Parameter 11 out of range"},
        {VI_ERROR_INSTR_PARAMETER12,  "ERROR: Parameter 12 out of range"},
        {VI_ERROR_INSTR_PARAMETER13,  "ERROR: Parameter 13 out of range"},
        {VI_ERROR_INSTR_PARAMETER14,  "ERROR: Parameter 14 out of range"},
        {VI_ERROR_INSTR_PARAMETER15,  "ERROR: Parameter 15 out of range"},
        {VI_ERROR_INSTR_PARAMETER16,  "ERROR: Parameter 16 out of range"},
        
        {VI_NULL, VI_NULL}
    };

    avq7760_status = viStatusDesc (instrSession, statusCode, errMessage);
    if (avq7760_status == VI_WARN_UNKNOWN_STATUS) {
        for (i = 0; statusDescArray[i].stringName; i++) {
            if (statusDescArray[i].stringVal == statusCode) {
                Fmt (errMessage, "%s<%s", statusDescArray[i].stringName);
                return (VI_SUCCESS);
            }
        }
        Fmt (errMessage, "%s<Unknown Error 0x%x[uw8p0]", statusCode);
        return (VI_WARN_UNKNOWN_STATUS);
    }
    
    avq7760_status = VI_SUCCESS;
    return avq7760_status;
}

/*===========================================================================*/
/* Function: Revision Query                                                  */
/* Purpose:  This function returns the driver and instrument revisions.      */
/*===========================================================================*/
ViStatus _VI_FUNC avq7760_Query_TypeRevision (ViSession instrSession,
                                               ViChar driverRev[], ViChar instrType[],
                                               ViChar instrCode[], ViChar instrRev[])
{
    ViStatus avq7760_status = VI_SUCCESS;
	ViChar rdBuffer[BUFFER_SIZE]; 
	ViUInt32 retCount = 0;     
	
    if ((avq7760_status = viWrite (instrSession, "*IDN?", 5, &retCount)) < 0)
        return avq7760_status;

    if ((avq7760_status = viRead (instrSession, rdBuffer, BUFFER_SIZE, &retCount)) < 0)
        return avq7760_status;

    if (Scan (rdBuffer, "%s>%s[dt44],%s[t44],%s[t44],%s", instrType, instrCode, instrRev) < 4)
        return VI_ERROR_INSTR_INTERPRETING_RESPONSE;

    Fmt (driverRev, "%s<%s", AVQ7760_REVISION);
    
    return avq7760_status;    
}

/*===========================================================================*/
/* Function: Close                                                           */
/* Purpose:  This function closes the instrument.                            */
/*===========================================================================*/
ViStatus CVIFUNC avq7760_close (ViSession instrSession)
{
    ViSession rmSession;
    ViStatus avq7760_status = VI_SUCCESS;

    if ((avq7760_status = viGetAttribute (instrSession, VI_ATTR_RM_SESSION, &rmSession)) < 0)
        return avq7760_status;
            
    avq7760_status = viClose (instrSession);
    viClose (rmSession);

    return avq7760_status;
}

/*****************************************************************************/
/*= UTILITY ROUTINES (Non-Exportable Functions) =============================*/
/*****************************************************************************/

/*===========================================================================*/
/* Function: Boolean Value Out Of Range - ViBoolean                          */
/* Purpose:  This function checks a Boolean to see if it is equal to VI_TRUE */
/*           or VI_FALSE. If the value is out of range, the return value is  */
/*           VI_TRUE, otherwise the return value is VI_FALSE.                */
/*===========================================================================*/
ViBoolean avq7760_invalidViBooleanRange (ViBoolean val)
{
    return ((val != VI_FALSE && val != VI_TRUE) ? VI_TRUE : VI_FALSE);
}

/*===========================================================================*/
/* Function: Short Signed Integer Value Out Of Range - ViInt16               */
/* Purpose:  This function checks a short signed integer value to see if it  */  
/*           lies between a minimum and maximum value.  If the value is out  */
/*           of range, the return value is VI_TRUE, otherwise the return     */
/*           value is VI_FALSE.                                              */
/*===========================================================================*/
ViBoolean avq7760_invalidViInt16Range (ViInt16 val, ViInt16 min, ViInt16 max)
{
    return ((val < min || val > max) ? VI_TRUE : VI_FALSE);
}

/*===========================================================================*/
/* Function: Long Signed Integer Value Out Of Range - ViInt32                */
/* Purpose:  This function checks a long signed integer value to see if it   */  
/*           lies between a minimum and maximum value.  If the value is out  */
/*           of range, the return value is VI_TRUE, otherwise the return     */
/*           value is VI_FALSE.                                              */
/*===========================================================================*/
ViBoolean avq7760_invalidViInt32Range (ViInt32 val, ViInt32 min, ViInt32 max)
{
    return ((val < min || val > max) ? VI_TRUE : VI_FALSE);
}

/*===========================================================================*/
/* Function: Unsigned Char Value Out Of Range - ViUInt8                      */
/* Purpose:  This function checks an unsigned char value to see if it        */  
/*           lies between a minimum and maximum value.  If the value is out  */
/*           of range, the return value is VI_TRUE, otherwise the return     */
/*           value is VI_FALSE.                                              */
/*===========================================================================*/
ViBoolean avq7760_invalidViUInt8Range (ViUInt8 val, ViUInt8 min, ViUInt8 max)
{
    return ((val < min || val > max) ? VI_TRUE : VI_FALSE);
}

/*===========================================================================*/
/* Function: Short Unsigned Integer Value Out Of Range - ViUInt16            */
/* Purpose:  This function checks a short unsigned integer value to see if it*/  
/*           lies between a minimum and maximum value.  If the value is out  */
/*           of range, the return value is VI_TRUE, otherwise the return     */
/*           value is VI_FALSE.                                              */
/*===========================================================================*/
ViBoolean avq7760_invalidViUInt16Range (ViUInt16 val, ViUInt16 min, ViUInt16 max)
{
    return ((val < min || val > max) ? VI_TRUE : VI_FALSE);
}

/*===========================================================================*/
/* Function: Long Unsigned Integer Value Out Of Range - ViUInt32             */
/* Purpose:  This function checks a long unsigned integer value to see if it */  
/*           lies between a minimum and maximum value.  If the value is out  */
/*           of range, the return value is VI_TRUE, otherwise the return     */
/*           value is VI_FALSE.                                              */
/*===========================================================================*/
ViBoolean avq7760_invalidViUInt32Range (ViUInt32 val, ViUInt32 min, ViUInt32 max)
{
    return ((val < min || val > max) ? VI_TRUE : VI_FALSE);
}

/*===========================================================================*/
/* Function: Real (Float) Value Out Of Range - ViReal32                      */
/* Purpose:  This function checks a real (float) value to see if it lies     */  
/*           between a minimum and maximum value.  If the value is out of    */
/*           range, the return value is VI_TRUE, otherwise the return value  */
/*           is VI_FALSE.                                                    */
/*===========================================================================*/
ViBoolean avq7760_invalidViReal32Range (ViReal32 val, ViReal32 min, ViReal32 max)
{
    return ((val < min || val > max) ? VI_TRUE : VI_FALSE);
}

/*===========================================================================*/
/* Function: Real (Double) Value Out Of Range - ViReal64                     */
/* Purpose:  This function checks a real (double) value to see if it lies    */  
/*           between a minimum and maximum value.  If the value is out of    */
/*           range, the return value is VI_TRUE, otherwise the return value  */
/*           is VI_FALSE.                                                    */
/*===========================================================================*/
ViBoolean avq7760_invalidViReal64Range (ViReal64 val, ViReal64 min, ViReal64 max)
{
    return ((val < min || val > max) ? VI_TRUE : VI_FALSE);
}


/*****************************************************************************/
/*----------- INSERT INSTRUMENT-DEPENDENT UTILITY ROUTINES HERE -------------*/
/*****************************************************************************/

/*===========================================================================*/
/* Function: Initialize Clean Up                                             */
/* Purpose:  This function is used only by the avq7760_Initialize function. */
/*			 When an error is detected this function is called to close the  */
/*           open resource manager and instrument object sessions and to     */
/*           set the instrSession that is returned from avq7760_Initialize  */
/*			 to VI_NULL.                                                     */
/*===========================================================================*/
ViStatus avq7760_initCleanUp (ViSession openRMSession,
                ViPSession openInstrSession, ViStatus currentStatus)
{
    viClose (*openInstrSession);
    viClose (openRMSession);
    *openInstrSession = VI_NULL;
    
    return currentStatus;
}

/*****************************************************************************/
/*=== END INSTRUMENT DRIVER SOURCE CODE =====================================*/
/*****************************************************************************/
